#include "StdAfx.h"
#include "DCMotor.h"
#include "queue.h"

///////////////////
// GLOBAL VARIABLES
bool stopDCMotorThread = false; // mechanism for stopping the worker thread
Queue *pDCMotorQ = NULL; // command queue from CDCMotor to the worker thread


///////////////////
// GLOBAL FUNCTIONS
DWORD WINAPI DCMotorWorkerThread(LPVOID param); // communications thread


//////////////////
// GLOBAL TYPEDEFS
typedef struct DCMotorCommParams { // parameters passed to the worker thread
	HWND hWnd; // for sending messages from the worker thread to the class
	CString csComPort;
} DCMotorCommParams;





///////////
// MESSAGES
int DCMOTOR_MESSAGE_ERROR = RegisterWindowMessage("DCMOTOR_MESSAGE_ERROR");
int DCMOTOR_MESSAGE_THREAD_FINISHED = RegisterWindowMessage("DCMOTOR_MESSAGE_THREAD_FINISHED");




// CStepperMotor

IMPLEMENT_DYNAMIC(CDCMotor, CStatic)


CDCMotor::CDCMotor(void)
: m_bOpen(false)
, m_bComsEstablished(false)
, m_csLastError(_T(""))
{
}

CDCMotor::~CDCMotor(void)
{
}

BEGIN_MESSAGE_MAP(CDCMotor, CStatic)
	ON_REGISTERED_MESSAGE(DCMOTOR_MESSAGE_THREAD_FINISHED, OnMessageThreadFinished)
	ON_REGISTERED_MESSAGE(DCMOTOR_MESSAGE_ERROR, OnMessageError)

END_MESSAGE_MAP()


//
int CDCMotor::Open(CString csPort)
{
	if (!m_bOpen){
		pDCMotorQ = new Queue(20);
		stopDCMotorThread = false;

		// start up the worker thread
		DWORD dwThreadID;
		DCMotorCommParams *pParams = new DCMotorCommParams;
		pParams->hWnd = m_hWnd;
		pParams->csComPort = csPort;
		m_hComThread = CreateThread(NULL, 0, DCMotorWorkerThread, pParams, 0, &dwThreadID);
		if (m_hComThread == NULL){
			return DCMOTOR_INVALID_HANDLE;
		}
		m_bOpen = true;
	}
	return 0;
}
//
int CDCMotor::Close(void)
{
	if (m_bOpen){
		stopDCMotorThread = true;

		// wait for the worker thread to finish
		DWORD result = WAIT_TIMEOUT;
		while (result != WAIT_OBJECT_0){
			result = WaitForSingleObject(m_hComThread, 200);
		}
	}
	m_bOpen = false;
	return 0;
}
//
// Message handler for errors that occur in the worker thread
afx_msg LRESULT CDCMotor::OnMessageError(WPARAM errnum, LPARAM l)
{
	int result = (int)errnum; 
	//if (result ==....
	m_csLastError = "An error occurred!"; 

	return 0;
}
//
// Message handler for notifying the class that the worker thread has finished
afx_msg LRESULT CDCMotor::OnMessageThreadFinished(WPARAM w, LPARAM l)
{
	int result = (int)w; 
	//if (result == ...){
	
	m_bOpen = false;
	m_bComsEstablished = false;
	pDCMotorQ->shutdown(); // properly disable the queue
	delete pDCMotorQ;

	return 0;
}
//


///////////////////
// THREAD FUNCTIONS
//
DWORD WINAPI DCMotorWorkerThread(LPVOID param){
	DCMotorCommParams *p = static_cast<DCMotorCommParams *>(param);

	// open the COM port
	CString portPath = "\\\\.\\" + p->csComPort;
	HANDLE hComm = CreateFile(
											portPath, 
											GENERIC_READ | GENERIC_WRITE,
											0, // required for serial comm
											0,
											OPEN_EXISTING, // required for serial comm
											0, // synchronous i/o
											0); // required for serial comm
	
	if (hComm == INVALID_HANDLE_VALUE){
		::PostMessage(p->hWnd, DCMOTOR_MESSAGE_THREAD_FINISHED, DCMOTOR_INVALID_HANDLE, 0);
		delete p;
		return DCMOTOR_INVALID_HANDLE;
	}

	// create a DCB structure for the serial port
	DCB oldDCB = {0}; // backup of old state
	if (!GetCommState(hComm, &oldDCB)){
		::PostMessage(p->hWnd, DCMOTOR_MESSAGE_THREAD_FINISHED, DCMOTOR_DCB_RETRIEVAL, 0);
		delete p;
		return DCMOTOR_DCB_RETRIEVAL;
	}
	DCB newDCB;
	FillMemory(&newDCB, sizeof(newDCB), 0);
	newDCB.DCBlength = sizeof(newDCB);
	if (!BuildCommDCB("9600,n,8,1", &newDCB)){ // using 38400 baud, no parity, 8 data bit, 1 stop bit
		::PostMessage(p->hWnd, DCMOTOR_MESSAGE_THREAD_FINISHED, DCMOTOR_DCB_CREATION, 0);
		delete p;
		return DCMOTOR_DCB_CREATION;
	}
	if (!SetCommState(hComm, &newDCB)){
		::PostMessage(p->hWnd, DCMOTOR_MESSAGE_THREAD_FINISHED, DCMOTOR_SET_COMM_STATE, 0);
		delete p;
		return DCMOTOR_SET_COMM_STATE;
	}

	// set timeout settings
	COMMTIMEOUTS timeouts;
	timeouts.ReadIntervalTimeout = 200;
	timeouts.ReadTotalTimeoutMultiplier = 10;
	timeouts.ReadTotalTimeoutConstant = 200;
	timeouts.WriteTotalTimeoutMultiplier = 10;
	timeouts.WriteTotalTimeoutConstant = 100;
	if (!SetCommTimeouts(hComm, &timeouts)){
		::PostMessage(p->hWnd, DCMOTOR_MESSAGE_THREAD_FINISHED, DCMOTOR_SET_TIMEOUTS, 0);
		delete p;
		return DCMOTOR_SET_TIMEOUTS;
	}

	// start communicating with the device
	try {
		while (!stopDCMotorThread){
			// attempt to get a queued command.  
			DCMotorCommand *command = (DCMotorCommand*)pDCMotorQ->RemoveHead();

			if (command != NULL){
				byte pData[7];
				pData[0] = 0x02;
				
				if (command->bForward)
					pData[1] = 1;
				else 
					pData[1] = 0;

				pData[2] = (byte)(command->nSpeed & 0x000000FF);
				pData[3] = (byte)((command->nSpeed & 0x0000FF00)>>8);
				pData[4] = (byte)((command->nSpeed & 0x00FF0000)>>16);
				pData[5] = (byte)((command->nSpeed & 0xFF000000)>>24);

				pData[6] = 0x03;

				// send out the data
				DWORD dwWritten;
				if (!WriteFile(hComm, pData, 7, &dwWritten, 0)){
					::PostMessage(p->hWnd, DCMOTOR_MESSAGE_THREAD_FINISHED, DCMOTOR_COM_WRITE, 0);
					return DCMOTOR_COM_WRITE;
				}
				else if (dwWritten != 7){
					::PostMessage(p->hWnd, DCMOTOR_MESSAGE_THREAD_FINISHED, DCMOTOR_NOT_ALL_WRITTEN, 0);
					return DCMOTOR_NOT_ALL_WRITTEN;
				}
				else { // done sending the command packet

				}
				delete command;
			}
			else { 
				// command was null, ignore it
			}
		} // end while
	} // end try
	catch (CTerminateThreadException *e){
		e->Delete();
	}

	if (hComm == INVALID_HANDLE_VALUE){
		::PostMessage(p->hWnd, DCMOTOR_MESSAGE_THREAD_FINISHED, DCMOTOR_INVALID_HANDLE, 0);	
		delete p;
		return DCMOTOR_INVALID_HANDLE;
	}
	else {
		if (CloseHandle(hComm))	{
			::PostMessage(p->hWnd, DCMOTOR_MESSAGE_THREAD_FINISHED, DCMOTOR_NO_ERRORS, 0);
			delete p;
			return DCMOTOR_NO_ERRORS;
		}
		else {
			::PostMessage(p->hWnd, DCMOTOR_MESSAGE_THREAD_FINISHED, DCMOTOR_HANDLE_NOT_CLOSED, 0);
			delete p;
			return DCMOTOR_HANDLE_NOT_CLOSED;
		}
	}
}
//
int CDCMotor::IssueCommand(bool bForward, unsigned int nSpeed)
{
	if (m_bOpen){		
		DCMotorCommand *pCommand = new DCMotorCommand;
		pCommand->bForward = bForward;
		pCommand->nSpeed = nSpeed;

		if (!pDCMotorQ->AddTail(pCommand)){
			delete pCommand;
			return DCMOTOR_QUEUE_FULL;
		}
	}

	return 0;
}
//
CString CDCMotor::GetLastError(void)
{
	return m_csLastError;
}
//


